/* DICElib (DIstributed CAVE Engine library) 
 * Copyright (c) 2001 Bruno Barberi Gnecco <brunobg@lsi.usp.br>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies of the Software, its documentation and marketing & publicity
 * materials, and acknowledgment shall be given in the documentation, materials
 * and software packages that this Software was used.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
 * IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

/* internal data */

#ifndef __DICELIB_H
#define __DICELIB_H

#ifdef __cplusplus
extern "C" {
#endif

/* undefine to use multicast */
//#define tcpipsync

#include <stdio.h>
#include <stdlib.h>
#include "dicelib.h"

#ifndef WIN32
#include <netinet/in.h>
#else
#include <winsock.h>
#endif

/*! \file _dicelib.h
  \brief Internal header
*/

/**
  \brief Debug macro.
  \param l verbose level.
  \param c any command.
*/
#define _dice_debug(l,c)		if ( l <= dice_data.verbose ) c

/**
  \brief Internal data
*/
struct _dice_data {
	int		verbose;		/**< debug level. */
	unsigned int	port;			/**< port to use. */
	int		max_connections;	/**< maximum listen connections. */
	FILE		*error;			/**< error. */
	void		*handle;		/**< dlopen handle to executable. */
	char		*color[2];		/**< escape code for colored output */
	int		original_only;		/**< only original can create, change or delete variables.*/

	/* server only */
	int		clientnum;		/**< Number of clients. */
	struct _clients {
		char		*address;	/**< client address */
		char		*path;		/**< name path */
		char		*function;	/**< individual function */
	} *clients;
	enum { NONE, RSH, SSH, BPROC } spawn; 	/**< spawning system */
	
	/* client only */
	int		id;			/**< client id */
	char		*controlname;		/**< control server name. */
	int		controlsock;		/**< socket to the control */
	int		(*function)(void *);    /**< individual function */

	/* multicast */
	struct in_addr	mcastaddr;		/**< multicast address */
	unsigned int	mcastport;		/**< multicast port */
	unsigned char	mcastTTL;		/**< multicast TTL */
	int		mcastsock;		/**< multicast socket */
	struct sockaddr_in mcastsockaddr;	/**< multicast address */
};
/**
  \brief Internal data structure declaration.
*/
extern struct _dice_data dice_data;

/**
  \brief Packet types.
*/
enum _dice_packet_type { 
	INVALID, 				/**< Invalid packet. */
	SYNC, 					/**< SYNC packet. */
	VARIABLE_ASYNC, 			/**< Asynchronous VARIABLE packet. */
	VARIABLE_SYNC, 				/**< Synchronous VARIABLE packet. */
	VARIABLE_FREE_ASYNC,			/**< Asynchronous FREE_VARIABLE packet. */
	VARIABLE_FREE_SYNC,			/**< Synchronous FREE_VARIABLE packet. */
	READY, 					/**< READY packet.  */
	CONTINUE, 				/**< CONTINUE packet. */
	HELLO,					/**< HELLO packet. */
	GOODBYE 				/**< GOODBYE packet. */
};
/**
  \brief Typedef encapsulation.
*/
typedef enum _dice_packet_type dice_packet_type;

/**
  \brief Packet structure.
*/
union _dice_packet {
	dice_packet_type	packtype;	/**< Packet type. */

	/** \brief SYNC packet */
	struct {
		dice_packet_type	packtype;	/**< Packet type. */
	} sync;

	/** \brief VARIABLE_* packet */
	struct {
		dice_packet_type	packtype;	/**< Packet type. */
		char			*name;		/**< Variable name. */
		char			*type;		/**< Variable type. */
		void			*value;		/**< Variable value. */
	} variable;
	
	/** \brief VARIABLE_FREE_* packet */
	struct {
		dice_packet_type	packtype;	/**< Packet type. */
		char			*name;		/**< Variable name. */
	} variable_free;

	/** \brief READY packet */
	struct {
		dice_packet_type	packtype;	/**< Packet type. */
	} ready;

	/** \brief CONT packet */
	struct {
		dice_packet_type	packtype;	/**< Packet type. */
	} cont;

	/** \brief HELLO packet */
	struct {
		dice_packet_type	packtype;	/**< Packet type. */
		unsigned long		id;		/**< Client ID. */
		char			*function;	/**< Client special function. */
	} hello;

	/** \brief GOODBYE packet */
	struct {
		dice_packet_type	packtype;	/**< Packet type. */
	} goodbye;
};
/**
  \brief Typedef encapsulation.
*/
typedef union _dice_packet dice_packet;

/** @name Internal functions
  */
/*@{*/
/* initialization */
extern int dice_type_init_defaults ( void );
extern void dice_type_free_all ( void );
extern int dice_sync_init ( void );
extern void dice_sync_end ( void );
extern int dice_async_init ( void );
extern void dice_async_end ( void );
extern int dice_init_client ( void );
extern char *defaultpath;

/* net */
extern void dice_sync_update_send ( void );
extern int dice_packet_send ( dice_packet *p, int sock );
extern dice_packet dice_packet_get ( int sock );
extern int dice_broadcast ( dice_packet *packet, int except );
extern int dice_control_loop ( int argc, char **argv );
extern int dice_multicast_send ( char *message );
extern int dice_multicast_get ( void );

/* utils */
extern int dice_type_register ( DICE_Type *t );
extern int dice_control_hierarchy ( int a, int b );
extern void dice_free ( void *data );
extern char *dice_packet_name ( dice_packet_type t );
extern void dice_printf(char *function, char *format, ...);
extern int dice_type_is_valid ( char *name );

#define dice_is_server()		(dice_data.clientnum != -1)
/*@}*/

#ifdef __cplusplus
}
#endif


#endif

